from typing import Optional, Union
from cnab.base.cnab_400 import CNAB400Registro1
from cnab.core.field import CNABField, CNABFieldType
from cnab.utils.dict_utils import set_if_has_value
from cnab.base.registro import Registro


class BradescoCnab400Registro1(CNAB400Registro1):
    _meta = {
        'tipo_registro': CNABField(
            length=1,
            default='1',
            validation=CNABFieldType.Int,
            required=True),
        'agencia_debito': CNABField(
            length=5,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'digito_agencia_debito': CNABField(
            length=1,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'razao_conta_corrente_pagador': CNABField(
            length=5,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'conta_corrente_pagador': CNABField(
            length=7,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'digito_conta_corrente_pagador': CNABField(
            length=1,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'filler0': CNABField(
            length=1,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'carteira_banco': CNABField(
            length=3,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'agencia': CNABField(
            length=5,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'conta': CNABField(
            length=7,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'conta_dv': CNABField(
            length=1,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'seu_numero': CNABField(
            length=25,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True),
        'codigo_banco_debito': CNABField(
            length=3,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'codigo_multa': CNABField(
            length=1,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'taxa_multa': CNABField(
            length=2,
            default='0',
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True),
        'nosso_numero': CNABField(
            length=11,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'nosso_numero_dv': CNABField(
            length=1,
            default='0',
            validation=CNABFieldType.Alfa,
            required=True),
        'vlr_bonificacao_dia': CNABField(
            length=8,
            default='0',
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True),
        'emissao_boleto': CNABField(
            length=1,
            default='2',
            validation=CNABFieldType.Int,
            required=True),
        'debito_automatico': CNABField(
            length=1,
            default='N',
            validation=CNABFieldType.Alfa,
            required=True),
        'filler6': CNABField(
            length=10,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True),
        'indicador_rateio': CNABField(
            length=1,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True),
        'endereco_aviso_debito': CNABField(# 2 = ignora
            length=1,
            default='2',
            validation=CNABFieldType.Int,
            required=True),
        'filler7': CNABField(
            length=2,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True),
        'codigo_movimento': CNABField(
            length=2,
            default=1,
            validation=CNABFieldType.Int,
            required=True),
        'numero_documento': CNABField(
            length=10,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True),
        'data_vencimento': CNABField(
            length=6,
            default='',
            validation=CNABFieldType.Date,
            required=True),
        'valor': CNABField(
            length=11,
            default='',
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True),
        'banco_cobrador': CNABField(
            length=3,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'agencia_cobradora': CNABField(
            length=5,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'especie_titulo': CNABField(
            length=2,
            default='1',
            validation=CNABFieldType.Int,
            required=True),
        'aceite': CNABField(
            length=1,
            default='N',
            validation=CNABFieldType.Alfa,
            required=True),
        'data_emissao': CNABField(
            length=6,
            default='',
            validation=CNABFieldType.Date,
            required=True),
        'cod_instrucao1': CNABField(
            length=2,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'cod_instrucao2': CNABField(
            length=2,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'vlr_juros': CNABField(
            length=11,
            default='0',
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True),
        'data_desconto': CNABField(
            length=6,
            default='0',
            validation=CNABFieldType.Date,
            required=True),
        'vlr_desconto': CNABField(
            length=11,
            default='0',
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True),
        'vlr_IOF': CNABField(
            length=11,
            default='0',
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True),
        'vlr_abatimento': CNABField(
            length=11,
            default='0',
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True),
        'tipo_inscricao': CNABField(
            length=2,
            default='',
            validation=CNABFieldType.Int,
            required=True),
        'numero_inscricao': CNABField(
            length=14,
            default='',
            validation=CNABFieldType.Int,
            required=True),
        'nome_pagador': CNABField(
            length=40,
            default='',
            validation=CNABFieldType.Alfa,
            required=True),
        'endereco_pagador': CNABField(
            length=40,
            default='',
            validation=CNABFieldType.Alfa,
            required=True),
        '1_mensagem': CNABField(
            length=12,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True),
        'cep_pagador': CNABField(
            length=8,
            default='',
            validation=CNABFieldType.Int,
            required=True),
        '2_mensagem': CNABField(
            length=60,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True),
        'numero_registro': CNABField(
            length=6,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
    }

    def __init__(self, header: Registro | None, parent: Registro | None, **kwargs: dict):
        super().__init__(header, parent, **kwargs)

    def inserir_mensagem(self, **kwargs):
        if not kwargs.get('mensagem'):
            return
        from .registro2 import BradescoCnab400Registro2
        return BradescoCnab400Registro2(self.header, self, kwargs)



from typing import Optional
from cnab.base.cnab_240 import CNAB240Registro3
from cnab.core.field import CNABField, CNABFieldType
from cnab.base.registro import Registro
from .registro3Q import BancoBrasil240Registro3Q
from .registro3R import BancoBrasil240Registro3R
from .registro3S1e2 import BancoBrasil240Registro3S1e2
from .registro3S3 import BancoBrasil240Registro3S3
from .registro1 import BancoBrasil240Registro1
from .registro0 import BancoBrasil240Registro0


class BancoBrasil240Registro3P(CNAB240Registro3):
    _meta = {
        "codigo_banco": CNABField(  # 1.3P
            length=3, default="001", validation=CNABFieldType.Int, required=True
        ),
        "codigo_lote": CNABField(  # 2.3P
            length=4, default=1, validation=CNABFieldType.Int, required=True
        ),
        "tipo_registro": CNABField(  # 3.3P
            length=1, default="3", validation=CNABFieldType.Int, required=True
        ),
        "numero_registro": CNABField(  # 4.3P
            length=5, default="0", validation=CNABFieldType.Int, required=True
        ),
        "seguimento": CNABField(  # 5.3P
            length=1, default="P", validation=CNABFieldType.Alfa, required=True
        ),
        "filler1": CNABField(  # 6.3P
            length=1, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "codigo_movimento": CNABField(  # 7.3P
            length=2,
            default="01",  # entrada de titulo
            validation=CNABFieldType.Int,
            required=True,
        ),
        # - ------------------ até aqui é igual para todo registro tipo 3
        "agencia": CNABField(  # 8.3P
            length=5, default="0", validation=CNABFieldType.Int, required=True
        ),
        "agencia_dv": CNABField(  # 9.3P
            length=1, default="", validation=CNABFieldType.Alfa, required=True
        ),
        "conta": CNABField(  # 10.3P
            length=12, default="0", validation=CNABFieldType.Int, required=True
        ),
        "conta_dv": CNABField(  # 11.3P
            length=1, default="0", validation=CNABFieldType.Alfa, required=True
        ),
        "filler2": CNABField(  # 12.3P
            length=1, default="0", validation=CNABFieldType.Alfa, required=True
        ),
        "nosso_numero": CNABField(  # 13.3P
            length=20, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "codigo_carteira": CNABField(  # 14.3P
            length=1, default="0", validation=CNABFieldType.Int, required=True
        ),
        "com_registro": CNABField(  # 15.3P
            length=1,
            default="1",  # combrança com registro
            validation=CNABFieldType.Int,
            required=True,
        ),
        "tipo_documento": CNABField(  # 16.3P
            length=1, default="2", validation=CNABFieldType.Int, required=True
        ),
        "emissao_boleto": CNABField(  # 17.3
            length=1, default="2", validation=CNABFieldType.Int, required=True
        ),
        "entrega_boleto": CNABField(  # 18.3P
            length=1,
            default="0",
            validation=CNABFieldType.Int,  # originalmente no manual esta alfa mas foi mudado para int para funcionar
            required=True,
        ),
        "seu_numero": CNABField(  # 19.3P   Campo de preenchimento obrigatório; preencher com Seu Número de controle do título
            length=15,
            default=" ",  # este espaço foi colocado para passa a validação para os seters do generico
            validation=CNABFieldType.Alfa,
            required=True,
        ),
        "data_vencimento": CNABField(  # 20.3
            length=8, default="", validation=CNABFieldType.Date, required=True
        ),
        "valor": CNABField(  # 21.3P
            length=13,
            default="0",
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True,
        ),
        "agencia_cobradora": CNABField(  # 22.3P
            length=5, default="0", validation=CNABFieldType.Int, required=True
        ),
        "agencia_cobradora_dv": CNABField(  # 23.3P
            length=1,
            default=" ",
            validation=CNABFieldType.Alfa,  # originalmente no manual esta alfa mas foi mudado para int para funcionar
            required=True,
        ),
        "especie_titulo": CNABField(  # 24.3P
            length=2, default="2", validation=CNABFieldType.Int, required=True
        ),
        "aceite": CNABField(  # 25.3P
            length=1, default="N", validation=CNABFieldType.Alfa, required=True
        ),
        "data_emissao": CNABField(  # 26.3P
            length=8, default="", validation=CNABFieldType.Date, required=True
        ),
        "codigo_juros": CNABField(  # 27.3P
            length=1, default="3", validation=CNABFieldType.Int, required=True
        ),
        "data_juros": CNABField(  # 28.3P
            length=8, default="0", validation=CNABFieldType.Date, required=True
        ),
        "vlr_juros": CNABField(  # 29.3P
            length=13,
            default="0",
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True,
        ),
        "codigo_desconto": CNABField(  # 30.3P
            length=1, default="0", validation=CNABFieldType.Int, required=True
        ),
        "data_desconto": CNABField(  # 31.3P
            length=8, default="0", validation=CNABFieldType.Date, required=True
        ),
        "vlr_desconto": CNABField(  # 32.3P
            length=13,
            default="0",
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True,
        ),
        "vlr_IOF": CNABField(  # 33.3P
            length=13,
            default="0",
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True,
        ),
        "vlr_abatimento": CNABField(  # 34.3P
            length=13,
            default="0",
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True,
        ),
        "seu_numero2": CNABField(  # 35.3P
            length=25, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "protestar": CNABField(  # 36.3P
            length=1, default="3", validation=CNABFieldType.Alfa, required=True
        ),
        "prazo_protesto": CNABField(  # 37.3P
            length=2, default="0", validation=CNABFieldType.Int, required=True
        ),
        "baixar": CNABField(  # 38.3P
            length=1, default="0", validation=CNABFieldType.Int, required=True
        ),
        "prazo_baixar": CNABField(  # 39.3P
            length=3, default="0", validation=CNABFieldType.Int, required=True
        ),
        "codigo_moeda": CNABField(  # 40.3P
            length=2, default="9", validation=CNABFieldType.Int, required=True
        ),
        "filler4": CNABField(  # 41.3P
            length=10, default="0", validation=CNABFieldType.Int, required=True
        ),
        "filler5": CNABField(  # 42.3P
            length=1, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
    }

    def _inserir_desconto_mensagem(self, **kwargs):
        desconto2 = kwargs.get("codigo_desconto2")
        desconto3 = kwargs.get("codigo_desconto3")
        mensagem = kwargs.get("mensagem")
        if not desconto2 and not desconto3 and not mensagem:
            return
        
        BancoBrasil240Registro3R(self.header, self, self.lote, **kwargs)

    def _inserir_detalhe(self, **kwargs: dict):
        BancoBrasil240Registro3Q(self.header, self, self.lote, **kwargs)
        self._inserir_desconto_mensagem(**kwargs)

        if int(kwargs.get("emissao_boleto")) != 1:
            return
        
        if kwargs.get('mensagem_frente'):
            kwargs['mensagem_140'] = kwargs.get('mensagem_frente')
            kwargs['tipo_impressao'] = 1
            BancoBrasil240Registro3S1e2(self.header, self, self.lote, **kwargs)

        if kwargs.get('mensagem_verso'):
            kwargs['mensagem_140'] = kwargs.get('mensagem_verso')
            kwargs['tipo_impressao'] = 2
            BancoBrasil240Registro3S1e2(self.header, self, self.lote, **kwargs)

        if not kwargs.get('mensagem_5') and not kwargs.get('mensagem_6') and not kwargs.get('mensagem_7') and not kwargs.get('mensagem_8'):
            return
        BancoBrasil240Registro3S3(self.header, self, self.lote, **kwargs)

    def __init__(
        self,
        header: Optional[BancoBrasil240Registro0],
        parent: Optional[Registro],
        lote: BancoBrasil240Registro1,
        **kwargs: dict,
    ):
        super().__init__(header, parent, lote, **kwargs)

        self._inserir_detalhe(**kwargs)



from typing import Optional, Union
from cnab.base.cnab_400 import CNAB400Registro9
from cnab.core.field import CNABField, CNABFieldType
from cnab.utils.dict_utils import set_if_has_value
from cnab.base.registro import Registro


class BradescoCnab400Registro9(CNAB400Registro9):
    _meta = {
        'tipo_registro': CNABField(
            length=1,
            default='9',
            validation=CNABFieldType.Int,
            required=True),
        'filler1': CNABField(            # 32.3P
            length=393,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True),
        'numero_registro': CNABField(       # 4.3R
            length=6,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
    }



from typing import Optional, Union
from cnab.base.cnab_400 import CNAB400Registro9
from cnab.core.field import CNABField, CNABFieldType
from cnab.utils.dict_utils import set_if_has_value
from cnab.base.registro import Registro


class ItauCnab400Registro9(CNAB400Registro9):
    _meta = {
        "tipo_registro": CNABField(
            length=1, default="9", validation=CNABFieldType.Int, required=True
        ),
        "filler1": CNABField(  # 32.3P
            length=393, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "numero_registro": CNABField(  # 4.3R
            length=6, default="0", validation=CNABFieldType.Int, required=True
        ),
    }



from cnab.base.registro import Registro
from cnab.base.registro_remessa import RegistroRemessa
from datetime import datetime
from .registro1 import CNAB240Registro1

class CNAB240Registro0(RegistroRemessa):
    registro1_class = None

    def __init__(
        self,
        **kwargs: dict
    ):
        self.counter = 0
        super().__init__(None, None, **kwargs)

    def get_data_geracao(self):
        return datetime.now()
    
    def get_hora_geracao(self):
        return datetime.now()
    
    def append(self, child: Registro):
        super().append(child)
        self.counter += 1

    def inserir_lote(self, **kwargs: dict) -> CNAB240Registro1:
        if not hasattr(self, 'registro1_class'):
            raise Exception("TODO: better exception here")
        if not self.registro1_class:
            raise Exception("TODO: better exception here")
        
        return self.registro1_class(self, self, **kwargs)



from typing import Union
from functools import reduce

class Modulo11Calculator:
    def get_factor(self, num):
        digits = [2, 3, 4, 5, 6, 7, 8, 9]
        if num >= len(digits):
            index = num % len(digits)
        else:
            index = num
        return digits[index]
    
    def compute_reversed_digits(self, num):
        num = str(num)
        digits = [int(digit) for digit in num]
        digits.reverse()
        return digits
    
    def compute_factors(self, reversed_digits):
        return [self.get_factor(i) for i in range(0, len(reversed_digits))]

    def compute_sum(self, reversed_digits, factors):
        return reduce(lambda prev, data: prev + (data[0] * data[1]), zip(reversed_digits, factors), 0)

    def compute_verify_digit(self, sum):
        value = 11 - (sum % 11)
        if value >= 10:
            return 0
        return value

    def __call__(self, num: Union[int, str]) -> int:
        digits = self.compute_reversed_digits(num)
        factors = self.compute_factors(digits)
        sum = self.compute_sum(digits, factors)
        return self.compute_verify_digit(sum)

compute_check_digit = Modulo11Calculator()



def format_decimal(value: float, field):
    print("#### ", field.name)

    value = "%.2f" % round(value, field.precision)
    if field.name == 'vlr_juros':
        print(value)
    value = value.replace(',', '').replace('.', '')
    if field.name == 'vlr_juros':
        print(value)
    value = value.rjust(field.length + field.precision, '0')
    if field.name == 'vlr_juros':
        print(len(value))

    return value




class CNABFieldValueRequiredError(Exception):
    def __init__(self, field_name: str):
        msg = f"Field not found or field with empty value: {field_name}"
        super().__init__(msg)

class CNABFieldNotIntegerError(Exception):
    def __init__(self, field_name: str):
        msg = f"The Integer field value is not parseable to int: {field_name}"
        super().__init__(msg)

class CNABFieldNotDecimalError(Exception):
    def __init__(self, field_name: str):
        msg = f"The decimal field value is not parseable to float: {field_name}"
        super().__init__(msg)

class CNABFieldTypeNotSupportedError(Exception):
    def __init__(self, field_type: str):
        msg = f"The field type {field_type} is not supported."
        super().__init__(msg)

class IsNotCNABFieldError(Exception):
    def __init__(self, field_name: str):
        msg = f"The meta field is not a CNABField instance: {field_name}"
        super().__init__(msg)

class IsNotValidDateTimeError(Exception):
    def __init__(self, field_name: str):
        msg = f"The date or time field needs an instance with strftime function: {field_name}"
        super().__init__(msg)



from typing import Optional
from datetime import datetime
from cnab.base.registro_remessa import RegistroRemessa
from cnab.base.registro import Registro
from cnab.core.enums import TipoServico, TipoInscricao
from cnab.core.exceptions import CNABInvalidTypeError
from cnab.core.especie import EspecieTitulo

class CNAB400Registro1(RegistroRemessa):
    def __init__(self, header: Optional["Registro"], parent: Optional["Registro"], **kwargs: dict):
        self.counter = 0
        super().__init__(header, parent, **kwargs)
        self.init_numero_registro()

    def init_numero_registro(self):
        self._data['numero_registro'] = self.header.counter
        self._data['numero_sequencial'] = self.header.counter

    def append(self, child: Registro):
        super().append(child)

        self.counter += 1

    def get_tipo_inscricao_empresa(self):
        return self.get_data_or_parent('tipo_inscricao_empresa')
    
    def get_numero_inscricao_empresa(self):
        return self.get_data_or_parent('numero_inscricao_empresa')
    
    def get_agencia(self):
        return self.get_data_or_parent('agencia')

    def get_conta(self):
        return self.get_data_or_parent('conta')

    def get_conta_dv(self):
        return self.get_data_or_parent('conta_dv')
    
    def get_data_emissao(self):
        return datetime.now()
    
    def get_numero_registro(self):
        return self.get_data_or_parent('numero_sequencial')

    def get_especie_titulo(self):
        field = self.get_field('codigo_banco')
        return EspecieTitulo.get_real_value(
            field.default,
            self._data.get('especie_titulo')
        )
    


from .registro0 import CNAB400Registro0
from .registro1 import CNAB400Registro1
from .registro2 import CNAB400Registro2
from .registro9 import CNAB400Registro9

__all__ = [
    'CNAB400Registro0',
    'CNAB400Registro1',
    'CNAB400Registro2',
    'CNAB400Registro9',
]



from .decimal import format_decimal
from .alfa import format_alfa
from .alfa2 import format_alfa2
from .integer import format_integer
from .date import format_date
from .time import format_time

__all__ = [
    'format_decimal',
    'format_alfa',
    'format_alfa2',
    'format_integer',
    'format_date',
    'format_time',
]



from typing import Optional, Union
from cnab.base.cnab_400 import CNAB400Registro0
from cnab.core.field import CNABField, CNABFieldType
from cnab.utils.dict_utils import set_if_has_value


class BradescoCnab400Registro0(CNAB400Registro0):
    _meta = {
        'identificacao_registro': CNABField(
            length=1,
            default='0',
            validation=CNABFieldType.Int,
            required=True),
        'operacao': CNABField(
            length=1,
            default='1',
            validation=CNABFieldType.Int,
            required=True),
        'literal_remessa': CNABField(
            length=7,
            default='remessa',
            validation=CNABFieldType.Alfa,
            required=True),
        'tipo_servico': CNABField(
            length=2,
            default='01',
            validation=CNABFieldType.Int,
            required=True),
        'literal_servico': CNABField(
            length=15,
            default='COBRANCA',
            validation=CNABFieldType.Alfa,
            required=True),
        'codigo_beneficiario': CNABField(
            length=20,
            default='',
            validation=CNABFieldType.Int,
            required=True),
        'nome_empresa': CNABField(
            length=30,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True),
        'codigo_banco': CNABField(
            length=3,
            default='237',
            validation=CNABFieldType.Int,
            required=True),
        'nome_banco': CNABField(
            length=15,
            default='Bradesco',
            validation=CNABFieldType.Alfa,
            required=True),
        'data_gravacao': CNABField(
            length=6,
            default='',# nao informar a data na instanciação - gerada dinamicamente
            validation=CNABFieldType.Date,
            required=True),
        'filler1': CNABField(
            length=8,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True),
        'identificacao_sistema': CNABField(
            length=2,
            default='MX',
            validation=CNABFieldType.Alfa,
            required=True),
        'numero_sequencial_arquivo': CNABField(
            length=7,
            default='1',
            validation=CNABFieldType.Int,
            required=True),
        'filler2': CNABField(
            length=277,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True),
        'numero_sequencial': CNABField(
            length=6,
            default='1',
            validation=CNABFieldType.Int,
            required=True),
    }

    def inserir_detalhe(self, **kwargs):
        from .registro1 import BradescoCnab400Registro1
        return BradescoCnab400Registro1(self, self, **kwargs)



from typing import Optional, Union
from cnab.base.cnab_400 import CNAB400Registro1
from cnab.core.field import CNABField, CNABFieldType
from cnab.utils.dict_utils import set_if_has_value
from cnab.base.registro import Registro
from .registro2 import ItauCnab400Registro2


class ItauCnab400Registro1(CNAB400Registro1):
    _meta = {
        "tipo_registro": CNABField(
            length=1, default="1", validation=CNABFieldType.Int, required=True
        ),
        "tipo_inscricao_empresa": CNABField(
            length=2, default="", validation=CNABFieldType.Int, required=True
        ),
        "numero_inscricao_empresa": CNABField(
            length=14, default="", validation=CNABFieldType.Int, required=True
        ),
        "agencia": CNABField(
            length=4, default="", validation=CNABFieldType.Int, required=True
        ),
        "filler1": CNABField(
            length=2, default="0", validation=CNABFieldType.Int, required=True
        ),
        "conta": CNABField(
            length=5, default="", validation=CNABFieldType.Int, required=True
        ),
        "conta_dv": CNABField(
            length=1, default="", validation=CNABFieldType.Int, required=True
        ),
        "filler2": CNABField(
            length=4, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "cod_intrucao": CNABField(
            length=4, default="0", validation=CNABFieldType.Int, required=True
        ),
        "seu_numero": CNABField(
            length=25, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "nosso_numero": CNABField(
            length=8, default="", validation=CNABFieldType.Int, required=True
        ),
        "qtd_moeda": CNABField(  # 34.3P
            length=8,
            default="0",
            validation=CNABFieldType.Decimal,
            precision=5,
            required=True,
        ),
        "carteira_banco": CNABField(  # 13.3P
            length=3, default="0", validation=CNABFieldType.Int, required=True
        ),
        "filler3": CNABField(
            length=21, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "cod_carteira": CNABField(  # 13.3P
            length=1, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "codigo_movimento": CNABField(  # codigo da ocorrencia no manual itau
            length=2,
            default="01",  # entrada de titulo
            validation=CNABFieldType.Int,
            required=True,
        ),
        "numero_documento": CNABField(
            length=10,
            default=" ",
            validation=CNABFieldType.Alfa,
            required=True,
        ),
        "data_vencimento": CNABField(  # 20.3
            length=6, default="", validation=CNABFieldType.Date, required=True
        ),
        "valor": CNABField(  # 21.3P
            length=11,
            default="",
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True,
        ),
        "codigo_banco": CNABField(
            length=3, default="341", validation=CNABFieldType.Int, required=True
        ),
        "agencia_cobradora": CNABField(  # 22.3P
            length=5, default="0", validation=CNABFieldType.Int, required=True
        ),
        "especie_titulo": CNABField(  # 24.3P
            length=2, default="2", validation=CNABFieldType.Int, required=True
        ),
        "aceite": CNABField(  # 25.3P
            length=1, default="N", validation=CNABFieldType.Alfa, required=True
        ),
        "data_emissao": CNABField(  # 26.3P
            length=6, default="", validation=CNABFieldType.Date, required=True
        ),
        "cod_instrucao1": CNABField(  # 24.3P
            length=2, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "cod_instrucao2": CNABField(  # 24.3P
            length=2, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "vlr_juros": CNABField(  # 29.3P
            length=11,
            default="0",
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True,
        ),
        "data_desconto": CNABField(  # 31.3P
            length=6, default="0", validation=CNABFieldType.Date, required=True
        ),
        "vlr_desconto": CNABField(  # 32.3P
            length=11,
            default="0",
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True,
        ),
        "vlr_IOF": CNABField(  # 33.3P
            length=11,
            default="0",
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True,
        ),
        "vlr_abatimento": CNABField(  # 34.3P
            length=11,
            default="0",
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True,
        ),
        "tipo_inscricao": CNABField(
            length=2, default="", validation=CNABFieldType.Int, required=True
        ),
        "numero_inscricao": CNABField(
            length=14, default="", validation=CNABFieldType.Int, required=True
        ),
        "nome_pagador": CNABField(  # 10.3Q
            length=30, default="", validation=CNABFieldType.Alfa, required=True
        ),
        "filler4": CNABField(
            length=10, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "endereco_pagador": CNABField(  # 11.3Q
            length=40, default="", validation=CNABFieldType.Alfa, required=True
        ),
        "bairro_pagador": CNABField(  # 12.3Q
            length=12, default="", validation=CNABFieldType.Alfa, required=True
        ),
        "cep_pagador": CNABField(  # 13.3Q
            length=8, default="", validation=CNABFieldType.Int, required=True
        ),
        "cidade_pagador": CNABField(  # 15.3Q
            length=15, default="", validation=CNABFieldType.Alfa, required=True
        ),
        "uf_pagador": CNABField(  # 16.3Q
            length=2,
            default="",  # combrança com registro
            validation=CNABFieldType.Alfa,
            required=True,
        ),
        "nome_avalista": CNABField(  # 18.3Q
            length=30, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "filler5": CNABField(
            length=4, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "data_mora": CNABField(  # 31.3P
            length=6, default="0", validation=CNABFieldType.Date, required=True
        ),
        "prazo_baixa": CNABField(  # 31.3P
            length=2, default="0", validation=CNABFieldType.Int, required=True
        ),
        "filler6": CNABField(
            length=1, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "numero_registro": CNABField(  # 4.3R
            length=6, default="0", validation=CNABFieldType.Int, required=True
        ),
    }

    def __init__(self, header: Optional[Registro], parent: Optional[Registro], **kwargs: dict):
        super().__init__(header, parent, **kwargs)
        self.inserir_multa(**kwargs)

    def inserir_multa(self, **kwargs: dict):
        if not kwargs.get('data_multa'):
            return
        ItauCnab400Registro2(self.header, self, **kwargs)

    def get_agencia_cobradora(self):
        return self.get_data_or_parent('agencia')



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

f = open(os.path.join(os.path.dirname(__file__), 'README.md'))
readme = f.read()
f.close()

setup(
    name='enhanced-cnab',
    version='0.0.9',
    description='Enhanced way to generate brazil banks CNAB files in python.',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Eduardo Oliveira',
    author_email='eduardo_y05@outlook.com',
    url='https://github.com/EduardoJM/enhanced-cnab',
    license='MIT',
    packages=find_packages(
        exclude = ("tests.*", "tests", )
    ),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    keywords='cnab,banks,boletos,remessa,lote,cnab240,cnab400',
)


from cnab.core.exceptions import CNABFieldNotDecimalError

def validate_decimal(value, field):
    try:
        value = float(str(value))
    except Exception:
        raise CNABFieldNotDecimalError(field.name)
    return value



class Remessa:
    banco: int
    layout: str
    kwargs: dict
    
    current_lote = 1
    end_line = '\r\n'
    

    def __init__(self, banco: int, layout: str, **kwargs: dict):
        self.banco = banco
        self.layout = layout
        self.kwargs = kwargs

    def append_lote(self, data: dict):
        # TODO: apenas se for 240
        #Criar Registro1
        self.current_lote = self.current_lote + 1
        return self



from typing import Optional, Union
from cnab.base.cnab_240 import CNAB240Registro0
from cnab.core.field import CNABField, CNABFieldType
from cnab.utils.dict_utils import set_if_has_value
from .registro1 import BancoBrasil240Registro1

class BancoBrasil240Registro0(CNAB240Registro0):
    registro1_class = BancoBrasil240Registro1
    _meta = {
        "codigo_banco": CNABField(  # 01.0
            length=3, default="001", validation=CNABFieldType.Int, required=True
        ),
        "codigo_lote": CNABField(  # 02.0
            length=4, default="0000", validation=CNABFieldType.Int, required=True
        ),
        "tipo_registro": CNABField(  # 03.0
            length=1, default="0", validation=CNABFieldType.Int, required=True
        ),
        "filler1": CNABField(  #  04.0
            length=9, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "tipo_inscricao": CNABField(  # 05.0
            length=1, default="0", validation=CNABFieldType.Int, required=True
        ),
        "numero_inscricao": CNABField(  # 06.0
            length=14, default="0", validation=CNABFieldType.Int, required=True
        ),
        "convenio": CNABField(  # 07.0 BB1
            length=9, default="0", validation=CNABFieldType.Int, required=True
        ),
        "cobranca_cedente": CNABField(  # 07.0 BB2
            length=4, default="0014", validation=CNABFieldType.Int, required=True
        ),
        "carteira_cobranca": CNABField(  # 07.0 BB3
            length=2, default="11", validation=CNABFieldType.Int, required=True
        ),
        "variacao_carteira_cobranca": CNABField(  # 07.0 BB4
            length=3, default="222", validation=CNABFieldType.Int, required=True
        ),
        "reservado_bb": CNABField(  # 07.0 BB5
            length=2, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "agencia": CNABField(  # 08.0
            length=5, default="0", validation=CNABFieldType.Int, required=True
        ),
        "agencia_dv": CNABField(  # 09.0
            length=1, default="", validation=CNABFieldType.Alfa, required=True
        ),
        "conta": CNABField(  # 10.0
            length=12, default="0", validation=CNABFieldType.Int, required=True
        ),
        "conta_dv": CNABField(  # 11.0
            length=1, default="0", validation=CNABFieldType.Alfa, required=True
        ),
        "filler2": CNABField(  # 12.0
            length=1, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "nome_empresa": CNABField(  # 13.0
            length=30, default="", validation=CNABFieldType.Alfa, required=True
        ),
        "nome_banco": CNABField(  # 14.0
            length=30,
            default="BANCO DO BRASIL S.A",
            validation=CNABFieldType.Alfa,
            required=True,
        ),
        "filler3": CNABField(  # 15.0
            length=10, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "codigo_remessa": CNABField(  # 16.0
            length=1, default="1", validation=CNABFieldType.Int, required=True
        ),
        "data_geracao": CNABField(  # 17.0
            length=8,
            default="",  # nao informar a data na instanciaÃ§Ã£o - gerada dinamicamente
            validation=CNABFieldType.Date,
            required=True,
        ),
        "hora_geracao": CNABField(  # 18.00
            length=6,
            default="",  # nao informar a data na instanciaÃ§Ã£o - gerada dinamicamente
            validation=CNABFieldType.Time,
            required=True,
        ),
        "numero_sequencial_arquivo": CNABField(  # 19.0
            length=6, default="0", validation=CNABFieldType.Int, required=True
        ),
        "versao_layout": CNABField(  # 20.0
            length=3, default="000", validation=CNABFieldType.Int, required=True
        ),
        "densidade_gravacao": CNABField(  # 21.0
            length=5, default="0", validation=CNABFieldType.Int, required=True
        ),
        "filler4": CNABField(  # 22.0
            length=20, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "situacao_arquivo2": CNABField(  # 23.0
            length=20, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "filler5": CNABField(
            length=14, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "filler6": CNABField(  # Caso a versão do leiaute seja 30,
            length=3,  # Deve ser preenchido com 'zeros' nas posições 226 a 228.
            default="000",
            validation=CNABFieldType.Alfa,
            required=True,
        ),
        "filler7": CNABField(
            length=12, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
    }

    def __init__(
        self,
        *,
        codigo_banco: Optional[Union[str, int]] = None,
        codigo_lote: Optional[Union[str, int]] = None,
        tipo_registro: Optional[Union[str, int]] = None,
        tipo_inscricao: Optional[Union[str, int]] = None,
        numero_inscricao: Optional[Union[str, int]] = None,
        convenio: Optional[Union[str, int]] = None,
        cobranca_cedente: Optional[Union[str, int]] = None,
        carteira_cobranca: Optional[Union[str, int]] = None,
        variacao_carteira_cobranca: Optional[Union[str, int]] = None,
        agencia: Optional[Union[str, int]] = None,
        agencia_dv: Optional[Union[str, int]] = None,
        conta: Optional[Union[str, int]] = None,
        conta_dv: Optional[Union[str, int]] = None,
        nome_empresa: Optional[str] = None,
        nome_banco: Optional[str] = None,
        codigo_remessa: Optional[Union[str, int]] = None,
        numero_sequencial_arquivo: Optional[Union[str, int]] = None,
        **kwargs
    ):
        set_if_has_value(kwargs, 'codigo_banco', codigo_banco)
        set_if_has_value(kwargs, 'codigo_lote', codigo_lote)
        set_if_has_value(kwargs, 'tipo_registro', tipo_registro)
        set_if_has_value(kwargs, 'tipo_inscricao', tipo_inscricao)
        set_if_has_value(kwargs, 'numero_inscricao', numero_inscricao)
        set_if_has_value(kwargs, 'convenio', convenio)
        set_if_has_value(kwargs, 'cobranca_cedente', cobranca_cedente)
        set_if_has_value(kwargs, 'carteira_cobranca', carteira_cobranca)
        set_if_has_value(kwargs, 'variacao_carteira_cobranca', variacao_carteira_cobranca)
        set_if_has_value(kwargs, 'agencia', agencia)
        set_if_has_value(kwargs, 'agencia_dv', agencia_dv)
        set_if_has_value(kwargs, 'conta', conta)
        set_if_has_value(kwargs, 'conta_dv', conta_dv)
        set_if_has_value(kwargs, 'nome_empresa', nome_empresa)
        set_if_has_value(kwargs, 'nome_banco', nome_banco)
        set_if_has_value(kwargs, 'codigo_remessa', codigo_remessa)
        set_if_has_value(kwargs, 'numero_sequencial_arquivo', numero_sequencial_arquivo)
        super().__init__(**kwargs)

    def inserir_lote(
        self,
        *,
        tipo_servico: Optional[Union[str, int]] = None,
        variacao: Optional[Union[str, int]] = None,
        **kwargs: dict
    ) -> BancoBrasil240Registro1:
        set_if_has_value(kwargs, 'tipo_servico', tipo_servico)
        set_if_has_value(kwargs, 'variacao', variacao)
        return super().inserir_lote(**kwargs)
